#!/bin/bash

set -e

SCRIPT_HOME=$(dirname "$0")
SCRIPT_HOME="`cd \"${SCRIPT_HOME}\" >/dev/null && pwd`"
SCRIPT_FOLDER="$(basename "$SCRIPT_HOME")"

usage()
{
  echo "Usage: `basename $0` command [options] [switches]"
  echo "  command:"
  echo "    download       [options]    [switches]"
  echo "    build          [options]    [switches]"
  echo "    start|run      [options]    [switches]"
  echo "    stop           [options]    [switches]"
  echo "    restart        [options]    [switches]"
  echo "    status         [options]    [switches]"
  echo "    cert           [options]    [switches]"
  echo "    dbms           [options]    [switches]"
  echo "    volume         [options]    [switches]"
  echo "    update         [options]    [switches]"
  echo "    scan           [options]    [switches]"
  echo "    diagnostic     [options]    [switches]"
  echo "    cleanup        [-image]     [-journal]"
  echo "    deploy         [options]    [switches]"
  echo "    submit         [options]    [switches]"
  echo "  options:"
  echo "    --c|--controller=<container name>  | use Controller container, default: $JS_CONTROLLER_IMAGE"
  echo "    --a|--agent=<container name>       | use Agent container, default: $JS_AGENT_IMAGE"
  echo "    --j|--joc=<container name>         | use JOC Cockpit container, default: $JS_JOC_IMAGE"
  echo "    --db=<container name>              | for build, run, dbms: use database container, default: $JS_DB_IMAGE"
  echo "    --versions=<number>                | for download: number of versions, default: $JS_DOWNLOAD_VERSIONS"
  echo "    --keypass=<password>               | for cert: keystore password, default: $JS_KEYSTORE_PASSWORD"
  echo "    --trustpass=<password>             | for cert: truststore password, default: $JS_TRUSTSTORE_PASSWORD"
  echo "    --volume=\"<vol1>\"[,\"<vol2>\"]       | for volume: specify volume shortcuts such as \"logs\", \"config\""
  echo "    --repeat=<number>                  | for diagnostic: repeat check in the given interval in seconds, default: $JS_DIAGNOSTIC_REPEAT"
  echo "    --url=<url>                        | for submit, use url for Controller web service, default: $JS_SUBMIT_URL"
  echo "    --dir=<path>                       | for submit, use directory for order files, default: $JS_SUBMIT_DIRECTORY"
  echo "    --file=<file|path>                 | for submit, use order file, relative to deploymenet directory or absolute path, default: $JS_SUBMIT_FILE"
  echo "    --id=<order id>                    | for submit, replacee order id from file with randomized id, default: $JS_SUBMIT_ID"
  echo "    --count=<number>                   | for submit, submit order for the given number of times, default: $JS_SUBMIT_COUNT"
  echo "  switches:"
  echo "    -c|-controller                     | use any Controller containers, default: $JS_CONTROLLER"
  echo "    -a|-agent                          | use any Agent containers, default: $JS_AGENT"
  echo "    -j|-joc                            | use any JOC Cockpit containers, default: $JS_JOC"
  echo "    -abort                             | for stop: abort running containers, default: $JS_STOP_ABORT"
  echo "    -kill                              | for stop: kill running containers, default: $JS_STOP_KILL"
  echo "    -image                             | for cleanup: remove dangling images, default: $JS_CLEANUP_IMAGE"
  echo "    -journal|-state                    | for build, start, restart, stop, update, cleanup: remove journal volume"
  echo "    -fd                                | for diagnostic: monitor file descriptors"
  echo "    -force                             | for volume: force volume replacement, for download: force web site download, default: $JS_FORCE"
}


diagnostic_file_descriptors()
{
# $1: controller, agent, joc
# $2: container name

  # echo ".. checking open file descriptors for container: $2"

  cid=`docker ps --no-trunc -aqf "name=$2"`

  if [ -z "$cid" ]
  then
    echo "unknonw container: $2"
    exit 1
  fi

  case "$1" in
    controller)  #
                 pid="`docker exec $2 ps -aux | awk '! /awk/ && /MasterMain/ { print $2}'`"
                 ;;
    agent)       # Alpine Linux with simplified ps command
                 pid="`docker exec $2 ps | awk '! /awk/ && /AgentMain/ { print $2}'`"
                 ;;
    joc)         #
                 pid="`docker exec $2 ps -aux | awk '! /awk/ && /jetty/ { print $2}'`"
                 ;;
  esac

  if [ -z "$pid" ]
  then
    echo "no process found in container: $2"
    exit 1
  fi
    
  fd="`docker exec -it $cid ls -1 /proc/$pid/fd | wc -l`"

  dt=`date '+%Y-%m-%d %H:%M:%S'`
  echo "# ---------- $dt ------------------------------------------- $2 --------------"
  echo "[$2][$pid][open file descriptors]$fd"
}

profile()
{
  if [ -f "$HOME/.js_profile" ]
  then
    echo ".. running profile: $HOME/.js_profile"

    if [ ! -x "$HOME/.js_profile" ]
    then
      chmod u+x $HOME/.js_profile
    fi

    . $HOME/.js_profile
  fi

  if [ -f "$HOME/.js_profile_$SCRIPT_FOLDER" ]
  then
    echo ".. running profile: $HOME/.js_profile_$SCRIPT_FOLDER"

    if [ ! -x "$HOME/.js_profile_$SCRIPT_FOLDER" ]
    then
      chmod u+x $HOME/.js_profile_$SCRIPT_FOLDER
    fi

    . $HOME/.js_profile_$SCRIPT_FOLDER
  else
    if [ -f "$SCRIPT_HOME/config/.js_profile" ]
    then
      echo ".. running profile: $SCRIPT_HOME/config/.js_profile"

      if [ ! -x "$SCRIPT_HOME/config/.js_profile" ]
      then
        chmod u+x $SCRIPT_HOME/config/.js_profile
      fi

      . $SCRIPT_HOME/config/.js_profile
    fi
  fi
}

profile_container()
(
  # $1: scope for controller, agent, joc, dbms
  # $2: container name

  if [ 1 = 0 ] && [ -f "$HOME/.js_profile_${SCRIPT_FOLDER}_$2" ]
  then
    echo ".. running profile: $HOME/.js_profile_${SCRIPT_FOLDER}_$2"

    if [ ! -x "$HOME/.js_profile_${SCRIPT_FOLDER}_$2" ]
    then
      chmod u+x $HOME/.js_profile_${SCRIPT_FOLDER}_$2
    fi

    . $HOME/.js_profile_${SCRIPT_FOLDER}_$2
  fi
)

profile_container_reset()
(
  # $1: scope for controller, agent, joc, dbms
  # $2: container name

  if [ 1 = 0 ] && [ -f "$HOME/.js_profile_${SCRIPT_FOLDER}_reset" ]
  then
    echo ".. running profile: $HOME/.js_profile_${SCRIPT_FOLDER}_reset"

    if [ ! -x "$HOME/.js_profile_${SCRIPT_FOLDER}_reset" ]
    then
      chmod u+x $HOME/.js_profile_${SCRIPT_FOLDER}_reset
    fi

    . $HOME/.js_profile_${SCRIPT_FOLDER}_reset
  fi
)

download()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern
  # $3: force: 0/1
  # $4: max. versions

  if [ "$3" = 1  ]
  then
    js_force="-force"
  else
    js_force=""
  fi

  if [ $4 -gt 1  ]
  then
    js_versions="--versions=$4"
  else
    js_versions=""
  fi

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -x $SCRIPT_HOME/$i/download.sh ]
                   then
                     echo ".. downloading Controller installer: $i"
                     $SCRIPT_HOME/$i/download.sh $js_force $js_versions
                   else
                     echo ".. skipping download of Controller installer: $i"
                   fi
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -x $SCRIPT_HOME/$i/download.sh ]
                   then
                     echo ".. downloading Agent installer: $i"
                     $SCRIPT_HOME/$i/download.sh $js_force $js_versions
                   else
                     echo ".. skipping download of Agent installer: $i"
                   fi
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -x $SCRIPT_HOME/$i/download.sh ]
                   then
                     echo ".. downloading JOC Cockpit installer: $i"
                     $SCRIPT_HOME/$i/download.sh $js_force $js_versions
                   else
                     echo ".. skipping download of JOC Cockpit installer: $i"
                   fi
               done
               ;;
  esac
}

scan_image()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -x $SCRIPT_HOME/$i/scan.sh ]
                   then
                     echo ".. scanning Controller image: $i"
                     $SCRIPT_HOME/$i/scan.sh
                   else
                     echo ".. skipping scan of Controller image: $i"
                   fi
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -x $SCRIPT_HOME/$i/scan.sh ]
                   then
                     echo ".. scanning Agent image: $i"
                     $SCRIPT_HOME/$i/scan.sh
                   else
                     echo ".. skipping scan of Agent image: $i"
                   fi
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -x $SCRIPT_HOME/$i/scan.sh ]
                   then
                     echo ".. scanning JOC Cockpit image: $i"
                     $SCRIPT_HOME/$i/scan.sh
                   else
                     echo ".. skipping scan of JOC Cockpit image: $i"
                   fi
               done
               ;;
  esac
}

build_image()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ -z "$c" ]
                   then
                     if [ -x $SCRIPT_HOME/$i/build.sh ]
                     then
                         echo ".. building Controller image: $i"
                         $SCRIPT_HOME/$i/build.sh
                     else
                         echo ".. skipping build of Controller image: $i"
                     fi
                   else
                     echo ".. Controller container running: $i"
                   fi
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ -z "$c" ]
                   then
                     if [ -x $SCRIPT_HOME/$i/build.sh ]
                     then
                         echo ".. building Agent image: $i"
                         $SCRIPT_HOME/$i/build.sh
                     else
                         echo ".. skipping build of Agent image: $i"
                     fi
                   else
                     echo ".. Agent container running: $i"
                   fi
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ -z "$c" ]
                   then
                     if [ -x $SCRIPT_HOME/$i/build.sh ]
                     then
                         echo ".. building JOC Cockpit image: $i"
                         $SCRIPT_HOME/$i/build.sh
                     else
                         echo ".. skipping build of JOC Cockpit image: $i"
                     fi
                   else
                     echo ".. JOC Cockpit container running: $i"
                   fi
               done
  esac
}

start_container()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ -z "$c" ]
                   then
		     profile_container controller $i
                     echo ".. starting Controller container: $i"
                     $SCRIPT_HOME/$i/run.sh
                     profile_container_reset controller $i
                   else
                     echo ".. Controller container running: $i"
                   fi
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f") 
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ -z "$c" ]
                   then
                     profile_container agent $i
                     echo ".. starting Agent container: $i"
                     $SCRIPT_HOME/$i/run.sh
                     profile_container_reset agent $i
                   else
                     echo ".. Agent container running: $i"
                   fi
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ -z "$c" ]
                   then
                     profile_container joc $1
                     echo ".. starting JOC container: $i"
                     $SCRIPT_HOME/$i/run.sh
                     profile_container_reset joc $i
                   else
                     echo ".. JOC container running: $i"
                   fi
               done
               ;;
      db)      #
               SCRIPT_CONTAINER="$(docker container ps -a -q --filter "name=$2")"
               if [ -z "$SCRIPT_CONTAINER" ]
               then
                 profil_container dbms $1
                 echo ".. starting database container: $2"
                 $SCRIPT_HOME/../dbms/$2/run.sh
                 profile_container_reset dbms $2
               else
                 echo ".. database container running: $2"
               fi
               ;;
  esac
}

restart_container()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping Controller container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi

                   profile_container controller $i
                   echo ".. starting Controller container: $i"
                   $SCRIPT_HOME/$i/run.sh
                   profile_container_reset controller $i
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping Agent container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi

                   profile_container agent $i
                   echo ".. starting Agent container: $i"
                   $SCRIPT_HOME/$i/run.sh
                   profile_container_reset agent $i
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping JOC container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi

                   profile_container joc $i
                   echo ".. starting JOC container: $i"
                   $SCRIPT_HOME/$i/run.sh
                   profile_container_reset joc $i
               done
               ;;
  esac
}

stop_container()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern
  # $3: abort
  # $4: kill

  if [ "$3" = "1" ]
  then
    JS_ACTION_ABORT="-abort"
  else
    JS_ACTION_ABORT=""
  fi

  if [ "$4" = "1" ]
  then
    JS_ACTION_KILL="-kill"
  else
    JS_ACTION_KILL=""
  fi

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping Controller container: $i"
                     $SCRIPT_HOME/$i/stop.sh $JS_ACTION_ABORT $JS_ACTION_KILL
                   else
                     echo "..Controller container not running: $i"
                   fi
               done
               ;;
      agent|director) #
               echo "director stop"
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping Agent container: $i"
                     $SCRIPT_HOME/$i/stop.sh $JS_ACTION_ABORT $JS_ACTION_KILL
                   else
                     echo ".. Agent container not running: $i"
                   fi
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping JOC container: $i"
                     $SCRIPT_HOME/$i/stop.sh $JS_ACTION_ABORT $JS_ACTION_KILL
                   else
                     echo ".. JOC container not running: $i"
                   fi
               done
               ;;
  esac
}

status_container()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. Controller container running: $i"
                   else
                     echo ".. Controller container not running: $i"
                   fi
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. Agent container running: $i"
                   else
                     echo ".. Agent container not running: $i"
                   fi
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. JOC container running: $i"
                   else
                     echo ".. JOC container not running: $i"
                   fi
               done
               ;;
  esac
}

create_controller_certificate()
{
  # $1: container path
  # $2: keystore password
  # $3: truststore password

  controller=$(basename "$1")

  keystore_file="$SCRIPT_HOME/$controller/build/private-https.p12"
  keystore_directory=$(dirname "$keystore_file")
  keystore_password="$2"

  truststore_file="$SCRIPT_HOME/config/certificates/controller-https.p12"
  truststore_directory=$(dirname "$truststore_file")
  truststore_password="$3"

  if [ -f "$keystore_file" ]; then
    hit="`keytool -list -v -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12" | grep $controller || true`"
    if [ ! "$hit" = "" ]
    then
      echo ".. removing alias from keystore: $controller"
      keytool -delete -alias "$controller" -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12" 2> /dev/null
    fi
  fi

  if [ -f "$truststore_file" ]; then
    hit="`keytool -list -v -keystore "$truststore_file" -keypass "$truststore_password" -storepass "$truststore_password" -storetype "PKCS12" | grep $controller || true`"
    if [ ! "$hit" = "" ]
    then
      echo ".. removing alias from truststore: $controller"
      keytool -delete -alias "$controller" -keystore "$truststore_file" -keypass "$truststore_password" -storepass "$truststore_password" -storetype "PKCS12"  2> /dev/null
    fi
  fi

  # generate Controller private key with alias name "<Controller Image Name>" in a keystore (private-https.p12)
  #   use the fully qualified hostname (FQDN) and name of your organization for the distinguished name
  echo ".. adding private key and public certificate to keystore"
  keytool -genkey -alias "$controller" -dname "CN=$controller,O=DEV,OU=SOS,DC=sos-berlin,DC=com" -validity 1461 -keyalg RSA -keysize 2048 -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12"
  keytool -list -keystore "$keystore_file" -storepass "$keystore_password" -storetype "PKCS12"

  # export the Controller public certificate to a file in PEM format (controller-https.cer)
  keytool -exportcert -rfc -noprompt -file "$keystore_directory/$controller.cer" -alias "$controller" -keystore "$keystore_file" -storepass "$keystore_password" -storetype "PKCS12"

  # import Controller public certificate to a truststore (agent-https.p12) by specifying the Agent public certificate file (agent-https.cer) and alias name (<Agent Image Name>)
  echo ".. adding public certificate to truststore"
  keytool -importcert -trustcacerts -noprompt -file "$keystore_directory/$controller.cer" -alias "$controller" -keystore "$truststore_file" -storepass "$truststore_password" -storetype "PKCS12"
  keytool -list -keystore "$truststore_file" -storepass "$truststore_password" -storetype "PKCS12"
}

create_agent_certificate()
{
  # $1: container path
  # $2: keystore password
  # $3: truststore password

  agent=$(basename "$1")

  keystore_file="$SCRIPT_HOME/$agent/build/private-https.p12"
  keystore_directory=$(dirname "$keystore_file")
  keystore_password="$2"

  truststore_file="$SCRIPT_HOME/config/certificates/agent-https.p12"
  truststore_directory=$(dirname "$truststore_file")
  truststore_password="$3"

  if [ -f "$keystore_file" ]; then
    hit="`keytool -list -v -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12" | grep $agent || true`"
    if [ ! "$hit" = "" ]
    then
      echo ".. removing alias from keystore: $agent"
      keytool -delete -alias "$agent" -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12" 2> /dev/null
    fi
  fi

  if [ -f "$truststore_file" ]; then
    hit="`keytool -list -v -keystore "$truststore_file" -keypass "$truststore_password" -storepass "$truststore_password" -storetype "PKCS12" | grep $agent || true`"
    if [ ! "$hit" = "" ]
    then
      echo ".. removing alias from trustore: $agent"
      keytool -delete -alias "$agent" -keystore "$truststore_file" -keypass "$truststore_password" -storepass "$truststore_password" -storetype "PKCS12" 2> /dev/null
    fi
  fi

  # generate Agent private key with alias name "<Agent Image Name>" in a keystore (private-https.p12)
  #   use the fully qualified hostname (FQDN) and name of your organization for the distinguished name
  echo ".. adding private key and public certificate to keystore"
  keytool -genkey -alias "$agent" -dname "CN=$agent,O=DEV,OU=SOS,DC=sos-berlin,DC=com" -validity 1461 -keyalg RSA -keysize 2048 -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12"
  keytool -list -keystore "$keystore_file" -storepass "$keystore_password" -storetype "PKCS12"

  # export the Agent public certificate to a file in PEM format (agent-https.cer)
  keytool -exportcert -rfc -noprompt -file "$keystore_directory/$agent.cer" -alias "$agent" -keystore "$keystore_file" -storepass "$keystore_password" -storetype "PKCS12"

  # import Agent public certificate to a truststore (agent-https.p12) by specifying the Agent public certificate file (agent-https.cer) and alias name (<Agent Image Name>)
  echo ".. adding public certificate to truststore"
  keytool -importcert -trustcacerts -noprompt -file "$keystore_directory/$agent.cer" -alias "$agent" -keystore "$truststore_file" -storepass "$truststore_password" -storetype "PKCS12"
  keytool -list -keystore "$truststore_file" -storepass "$truststore_password" -storetype "PKCS12"
}

create_joc_certificate()
{
  # $1: container path
  # $2: keystore password
  # $3: truststore password

  joc=$(basename "$1")

  keystore_file="$SCRIPT_HOME/$joc/build/private-https.p12"
  keystore_directory=$(dirname "$keystore_file")
  keystore_password="$2"

  truststore_file="$SCRIPT_HOME/config/certificates/joc-https.p12"
  truststore_directory=$(dirname "$truststore_file")
  truststore_password="$3"

  if [ -f "$keystore_file" ]; then
    hit="`keytool -list -v -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12" | grep $joc || true`"
    if [ ! "$hit" = "" ]
    then
      echo ".. removing alias from keystore: $joc"
      keytool -delete -alias "$joc" -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12" 2> /dev/null
    fi
  fi

  if [ -f "$truststore_file" ]; then
    hit="`keytool -list -v -keystore "$truststore_file" -keypass "$truststore_password" -storepass "$truststore_password" -storetype "PKCS12" | grep $joc || true`"
    if [ ! "$hit" = "" ]
    then
      echo ".. removing alias from keystore: $joc"
      keytool -delete -alias "$joc" -keystore "$truststore_file" -keypass "$truststore_password" -storepass "$truststore_password" -storetype "PKCS12" 2> /dev/null
    fi
  fi

  # generate JOC private key with alias name "<JOC Image Name>" in a keystore (private-https.p12)
  #   use the fully qualified hostname (FQDN) and name of your organization for the distinguished name
  echo ".. adding private key and public certificate to keystore"
  keytool -genkey -alias "$joc" -dname "CN=$joc,O=DEV,OU=SOS,DC=sos-berlin,DC=com" -validity 1461 -keyalg RSA -keysize 2048 -keystore "$keystore_file" -keypass "$keystore_password" -storepass "$keystore_password" -storetype "PKCS12"
  keytool -list -keystore "$keystore_file" -storepass "$keystore_password" -storetype "PKCS12"

  # export the JOC public certificate to a file in PEM format (joc-https.cer)
  keytool -exportcert -rfc -noprompt -file "$keystore_directory/$joc.cer" -alias "$joc" -keystore "$keystore_file" -storepass "$keystore_password" -storetype "PKCS12"

  # import JOC public certificate to a truststore (joc-https.p12) by specifying the Agent public certificate file (joc-https.cer) and alias name (<JOC Image Name>)
  echo ".. adding public certificate to truststore"
  keytool -importcert -trustcacerts -noprompt -file "$keystore_directory/$joc.cer" -alias "$joc" -keystore "$truststore_file" -storepass "$truststore_password" -storetype "PKCS12"
  keytool -list -keystore "$truststore_file" -storepass "$truststore_password" -storetype "PKCS12"
}

create_certificate()
{
  # $1: scope for controller, agent, joc
  # $2: image
  # $3: keystore password
  # $4: truststore password

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   echo ".. creating certificate for Controller container: $f"
                   create_controller_certificate "$f" "$3" "$4"
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   echo ".. creating certificate for Agent container: $f"
                   create_agent_certificate "$f" "$3" "$4"
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   echo ".. creating certificate for JOC container: $f"
                   create_joc_certificate "$f" "$3" "$4"
               done
               ;;
  esac 
}

switch_dbms()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping Controller container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi
                   echo ".. switching DBMS for Controller container: $i"
                   $SCRIPT_HOME/$i/dbms.sh $3
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping JOC container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi
                   echo ".. switching DBMS for JOC container: $i"
                   $SCRIPT_HOME/$i/dbms.sh $3
               done
               ;;
  esac
}

create_volume()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern
  # $3: list of volumes
  # $4: 1/0 for -force

  if [ ! -z "$3" ]
  then
    js_vol=$3
  else
    js_vol=""
  fi

  if [ "$4" = "1" ]
  then
    js_force="-force"
  else
    js_force=""
  fi

  case "$1" in
  controller)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping Controller container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi
                   echo ".. creating volume for Controller container: $i"
                   $SCRIPT_HOME/$i/volume.sh $js_force $js_vol
               done
               ;;
 agent|director) #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping Agent container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi
                   echo ".. creating volume for Agent container: $i"
                   $SCRIPT_HOME/$i/volume.sh $js_force $js_vol
               done
               ;;
      joc)     #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   c=$(docker container ps -a -q --filter "name=$i")
                   if [ ! -z "$c" ]
                   then
                     echo ".. stopping JOC container: $i"
                     $SCRIPT_HOME/$i/stop.sh
                   fi
                   echo ".. creating volume for JOC container: $i"
                   $SCRIPT_HOME/$i/volume.sh $js_force $js_vol
               done
               ;;
  esac
}

cleanup_image()
{
  instances=$(docker images --quiet --filter "dangling=true")
  for f in $instances; do
    docker image rm $f
  done
}

cleanup_logs()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
           *)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -d "$SCRIPT_HOME/$i/logs" ]
                   then
                     echo ".. removing log files for container: $i"
                     rm -frd $SCRIPT_HOME/$i/logs/*
                   fi

                   # if [ -d "$SCRIPT_HOME/$i/logs/history" ]
                   # then
                   #   echo ".. removing history files for container: $i"
                   #   rm -f $SCRIPT_HOME/$i/logs/history/*
                   # fi

               done
               ;;
  esac
}

cleanup_journal()
{
  allInstancesAreStopped=1

  instances=$(ls -d $SCRIPT_HOME/* | egrep ".*(controller|agent)-.*")
  for f in $instances; do
    i=$(basename "$f")
    c=$(docker container ps -a -q --filter "name=$i")
    if [ ! -z "$c" ] && [ "${allInstancesAreStopped}" = "1" ]
    then
      echo ".. container running: $i"
      allInstancesAreStopped=0
    fi
  done

  # do not remove any Agent or Controller journal if its cntainer is running
  if [ ${allInstancesAreStopped} = 1 ]
  then
    for f in $instances; do
      i=$(basename "$f")
      echo ".. removing journal for container: $i"
      # $SCRIPT_HOME/$i/volume.sh -force --vol=state
      rm -frd $SCRIPT_HOME/$i/var/state/*
    done
  else
    echo ".. cannot remove journal of Controllers and Agents: containers running"
  fi
}

deploy()
{
  # $1: scope for controller, agent, joc
  # $2: container name or pattern

  case "$1" in
           *)  #
               instances=$(ls -d $SCRIPT_HOME/$2 2>/dev/null || :)
               if [ -z "$instances" ]; then return 0; fi
               for f in $instances; do
                   i=$(basename "$f")
                   if [ -f "$SCRIPT_HOME/$i/deploy.sh" ]
                   then
                     echo ".. deploying to container: $i"
                     $SCRIPT_HOME/$i/deploy.sh
                   fi
               done
               ;;
  esac
}

submit_order()
{
  # $1: directory
  # $2: file
  # $3: id
  # $4: count
  # $5: url

  count=0
  while [ $count -lt $4 ]; do
    if [ -z "$2" ]
    then
      # no file specified, use directory only
      files=(`ls -A $1/`)
      directory="$1/"
    else
      if [ -z "$1" ]
      then
        # no directory specified, use file only
        files=(`ls -A $2`)
        directory=""
      else
        files=(`ls -A $1/$2`)
        directory=""
      fi
    fi

    for f in $files; do
      file=${directory}${f}
      if [ ! -z "$3" ]
      then
        inputfile=$(mktemp /tmp/submit.XXXXXX)
        sed "s/$3/$RANDOM/g" $file > $inputfile
      else
        inputfile=$file
      fi

      echo "+------------------------------+"
      echo ".. submitting file: $inputfile"
      cat $inputfile
      echo "+------------------------------+"

      curl --insecure --silent --show-error --request POST --data "@$inputfile" --include --max-time 15 --header "Accept: application/json" --header "Content-Type: application/json" $5/controller/api/order
      echo ""
    done

    let count=count+1
 done
}

##########################
# profile processing
profile

##########################
# set options
JS_CONTROLLER_IMAGE=${JS_DEFAULT_CONTROLLER_IMAGE:-*controller*}
JS_AGENT_IMAGE=${JS_DEFAULT_AGENT_IMAGE:-*agent*}
JS_JOC_IMAGE=${JS_DEFAULT_JOC_IMAGE:-*joc*}

JS_CONTROLLER=${JS_DEFAULT_CONTROLLER:-0}
JS_AGENT=${JS_DEFAULT_AGENT:-0}
JS_JOC=${JS_DEFAULT_JOC:-0}

JS_DOWNLOAD_VERSIONS=${JS_DEFAULT_DOWNLOAD_VERSIONS:-1}
JS_KEYSTORE_PASSWORD=${JS_DEFAULT_KEYSTORE_PASSWORD:-jobscheduler}
JS_TRUSTSTORE_PASSWORD=${JS_DEFAULT_TRUSTSTORE_PASSWORD:-jobscheduler}

JS_STOP_ABORT=${JS_DEFAULT_STOP_ABORT:-0}
JS_STOP_KILL=${JS_DEFAULT_STOP_KILL:-0}
JS_CLEANUP_IMAGE=${JS_DEFAULT_CLEANUP_IMAGE:-0}
JS_CLEANUP_JOURNAL=${JS_DEFAULT_CLEANUP_JOURNAL:-0}
JS_DB_IMAGE=${JS_DEFAULT_DB_IMAGE}
JS_FORCE=${JS_DEFAULT_FORCE:-0}
JS_VOLUMES=${JS_DEFAULT_VOLUME:-""}
JS_DIAGNOSTIC_FILE_DESCRIPTORS=${JS_DEFAULT_DIAGNOSTIC_FILE_DESCRIPTORS:-1}
JS_DIAGNOSTIC_REPEAT=${JS_DEFAULT_DIAGNOSTIC_REPEAT:-10}

JS_SUBMIT_URL=${JS_DEFAULT_SUBMIT_URL:-http://localhost:4444}
JS_SUBMIT_DIRECTOY=${JS_DEFAULT_SUBMIT_DIRECTORY}
JS_SUBMIT_FILE=${JS_DEFAULT_SUBMIT_FILE}
JS_SUBMIT_ID=${JS_DEFAULT_SUBMIT_ID:-UNIQUE-ORDER-ID}
JS_SUBMIT_COUNT=${JS_DEFAULT_SUBMIT_COUNT:-1}

for option in "$@"
do
  case "$option" in
         --agent=*)           JS_AGENT=1 && JS_AGENT_IMAGE=`echo "$option" | sed 's/--agent=//'`
                              ;;
         --a=*)               JS_AGENT=1 && JS_AGENT_IMAGE=`echo "$option" | sed 's/--a=//'`
                              ;;
         --controller=*)      JS_CONTROLLER=1 && JS_CONTROLLER_IMAGE=`echo "$option" | sed 's/--controller=//'`
                              ;;
         --c=*)               JS_CONTROLLER=1 && JS_CONTROLLER_IMAGE=`echo "$option" | sed 's/--c=//'`
                              ;;
         --joc=*)             JS_JOC=1 && JS_JOC_IMAGE=`echo "$option" | sed 's/--joc=//'`
                              ;;
         --j=*)               JS_JOC=1 && JS_JOC_IMAGE=`echo "$option" | sed 's/--j=//'`
                              ;;
         --db=*)              JS_DB_IMAGE=`echo "$option" | sed 's/--db=//'`
                              ;;
         --versions=*)        JS_DOWNLOAD_VERSIONS=`echo "$option" | sed 's/--versions=//'`
                              ;;
         --keypass=*)         JS_KEYSTORE_PASSWORD=`echo "$option" | sed 's/--keypass=//'`
                              ;;
         --trustpass=*)       JS_TRUSTSTORE_PASSWORD=`echo "$option" | sed 's/--trustpass=//'`
                              ;;
         --url=*)             JS_SUBMIT_URL=`echo "$option" | sed 's/--url=//'`
                              ;;
         --dir=*)             JS_SUBMIT_DIRECTORY=`echo "$option" | sed 's/--dir=//'`
                              ;;
         --file=*)            JS_SUBMIT_FILE=`echo "$option" | sed 's/--file=//'`
                              ;;
         --id=*)              JS_SUBMIT_ID=`echo "$option" | sed 's/--id=//'`
                              ;;
         --count=*)           JS_SUBMIT_COUNT=`echo "$option" | sed 's/--count=//'`
                              ;;
         --volumes=*)         JS_VOLUMES=`echo "$option" | sed 's/--volumes=//'`
                              ;;
         --repeat=*)          JS_DIAGNOSTIC_REPEAT=`echo "$option" | sed 's/--repeat=//'`
                              ;;
         -c|-controller)      JS_CONTROLLER=1
                              ;;
         -a|-agent)           JS_AGENT=1
                              ;;
         -j|-joc)             JS_JOC=1
                              ;;
         -abort*)             JS_STOP_ABORT=1
                              ;;
         -kill*)              JS_STOP_KILL=1
                              ;;
         -image*)             JS_CLEANUP_IMAGE=1
                              ;;
         -journal*|-state*)   JS_CLEANUP_JOURNAL=1
                              ;;
         -fd*)                JS_DIAGNOSTIC_FILE_DESCRIPTORS=1
                              ;;
         -force*)             JS_FORCE=1
                              ;;
         download|build|start|run|stop|restart|status|cert|dbms|volume|update|scan|diagnostic|cleanup|deploy|submit*)
                              ;;
         *)                   echo "unknown argument: $option"
                              usage
                              exit 1
                              ;;
  esac
done


##########################
# ... and action
case "$1" in
    download)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          download controller "$JS_CONTROLLER_IMAGE" "$JS_FORCE" "$JS_DOWNLOAD_VERSIONS"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          download agent "$JS_AGENT_IMAGE" "$JS_FORCE" "$JS_DOWNLOAD_VERSIONS"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          download joc "$JS_JOC_IMAGE" "$JS_FORCE" "$JS_DOWNLOAD_VERSIONS"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          download controller "$JS_CONTROLLER_IMAGE" "$JS_FORCE" "$JS_DOWNLOAD_VERSIONS"
          download agent      "$JS_AGENT_IMAGE"      "$JS_FORCE" "$JS_DOWNLOAD_VERSIONS"
          download joc        "$JS_JOC_IMAGE"        "$JS_FORCE" "$JS_DOWNLOAD_VERSIONS"
        fi
        ;;
    build)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          stop_container controller "$JS_CONTROLLER_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          stop_container agent "$JS_AGENT_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          stop_container joc "$JS_JOC_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          stop_container controller "$JS_CONTROLLER_IMAGE"
          stop_container agent      "$JS_AGENT_IMAGE"
          stop_container joc        "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CLEANUP_JOURNAL}" = 1 ]
        then
          cleanup_journal
        fi

        if [ "${JS_CONTROLLER}" = 1 ]
        then
          build_image controller "$JS_CONTROLLER_IMAGE"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          build_image agent "$JS_AGENT_IMAGE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          build_image joc "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          build_image controller "$JS_CONTROLLER_IMAGE"
          build_image agent      "$JS_AGENT_IMAGE"
          build_image joc        "$JS_JOC_IMAGE"
        fi
        ;;
    start|run)
        if [ "${JS_CLEANUP_JOURNAL}" = 1 ]
        then
          cleanup_journal
        fi

        if [ "${JS_CONTROLLER}" = 1 ]
        then
          start_container controller "$JS_CONTROLLER_IMAGE"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          start_container agent "$JS_AGENT_IMAGE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          start_container joc "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          start_container controller "$JS_CONTROLLER_IMAGE"
          start_container agent      "$JS_AGENT_IMAGE"
          start_container joc        "$JS_JOC_IMAGE"
        fi
        ;;
    restart)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          stop_container controller "$JS_CONTROLLER_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          stop_container agent "$JS_AGENT_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_CLEANUP_JOURNAL}" = 1 ]
        then
          cleanup_journal
        fi

        if [ "${JS_CONTROLLER}" = 1 ]
        then
          start_container controller "$JS_CONTROLLER_IMAGE"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          start_container agent "$JS_AGENT_IMAGE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          restart_container joc "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          stop_container controller "$JS_CONTROLLER_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
          stop_container agent      "$JS_AGENT_IMAGE"      "$JS_STOP_ABORT" "$JS_STOP_KILL"

          if [ "${JS_CLEANUP_JOURNAL}" = 1 ]
          then
            cleanup_journal
          fi

          start_container controller "$JS_CONTROLLER_IMAGE"
          start_container agent      "$JS_AGENT_IMAGE"
          restart_container joc      "$JS_JOC_IMAGE"
        fi
        ;;
    stop)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          stop_container controller "$JS_CONTROLLER_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          stop_container agent "$JS_AGENT_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          stop_container joc "$JS_JOC_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          stop_container controller "$JS_CONTROLLER_IMAGE" "$JS_STOP_ABORT" "$JS_STOP_KILL"
          stop_container agent      "$JS_AGENT_IMAGE"      "$JS_STOP_ABORT" "$JS_STOP_KILL"
          stop_container joc        "$JS_JOC_IMAGE"        "$JS_STOP_ABORT" "$JS_STOP_KILL"
        fi

        if [ "${JS_CLEANUP_JOURNAL}" = 1 ]
        then
          cleanup_journal
        fi
        ;;
    status)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          status_container controller "$JS_CONTROLLER_IMAGE"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          status_container agent "$JS_AGENT_IMAGE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          status_container joc "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          status_container controller "$JS_CONTROLLER_IMAGE"
          status_container agent      "$JS_AGENT_IMAGE"
          status_container joc        "$JS_JOC_IMAGE"
        fi
        ;;
    cert)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          create_certificate controller "$JS_CONTROLLER_IMAGE" "$JS_KEYSTORE_PASSWORD" "$JS_TRUSTSTORE_PASSWORD"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          create_certificate agent "$JS_AGENT_IMAGE" "$JS_KEYSTORE_PASSWORD" "$JS_TRUSTSTORE_PASSWORD"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          create_certificate joc "$JS_JOC_IMAGE" "$JS_KEYSTORE_PASSWORD" "$JS_TRUSTSTORE_PASSWORD"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          create_certificate controller "$JS_CONTROLLER_IMAGE" "$JS_KEYSTORE_PASSWORD" "$JS_TRUSTSTORE_PASSWORD"
          create_certificate agent      "$JS_AGENT_IMAGE"      "$JS_KEYSTORE_PASSWORD" "$JS_TRUSTSTORE_PASSWORD"
          create_certificate joc        "$JS_JOC_IMAGE"        "$JS_KEYSTORE_PASSWORD" "$JS_TRUSTSTORE_PASSWORD"
        fi
        ;;
    dbms)
        if [ -z "${JS_DB_IMAGE}" ]
        then
          echo ".. use option --db=<image name> to specify the target DBMS"
          exit 1
        fi

        if [ "${JS_CONTROLLER}" = 1 ]
        then
          switch_dbms controller "$JS_CONTROLLER_IMAGE" "$JS_DB_IMAGE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
         switch_ dbms joc "$JS_JOC_IMAGE" "$JS_DB_IMAGE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          switch_dbms controller "$JS_CONTROLLER_IMAGE" "$JS_DB_IMAGE"
          switch_dbms joc        "$JS_JOC_IMAGE"        "$JS_DB_IMAGE"
        fi
        ;;
    volume)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          create_volume controller "$JS_CONTROLLER_IMAGE" "$JS_VOLUMES" "$JS_FORCE"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          create_volume  agent "$JS_AGENT_IMAGE" "$JS_VOLUMES" "$JS_FORCE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          create_volume joc "$JS_JOC_IMAGE" "$JS_VOLUMES" "$JS_FORCE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          create_volume controller "$JS_CONTROLLER_IMAGE" "$JS_VOLUMES" "$JS_FORCE"
          create_volume agent      "$JS_AGENT_IMAGE"      "$JS_VOLUMES" "$JS_FORCE"
          create_volume joc        "$JS_JOC_IMAGE"        "$JS_VOLUMES" "$JS_FORCE"
        fi
        ;;
    update)
        JS_ARGS="`echo $@ | sed -e "s/update/stop/g"`"
        /bin/sh -c "$0 $JS_ARGS"
        JS_ARGS="`echo $@ | sed -e "s/update/download/g"`"
        /bin/sh -c "$0 $JS_ARGS"
        JS_ARGS="`echo $@ | sed -e "s/update/build/g"`"
        /bin/sh -c "$0 $JS_ARGS"
        JS_ARGS="`echo $@ | sed -e "s/update/start/g"`"
        /bin/sh -c "$0 $JS_ARGS"
        ;;
    scan)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          scan_image controller "$JS_CONTROLLER_IMAGE"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          scan_image  agent "$JS_AGENT_IMAGE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          scan_image joc "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          scan_image  controller "$JS_CONTROLLER_IMAGE"
          scan_image  agent      "$JS_AGENT_IMAGE"
          scan_image  joc        "$JS_JOC_IMAGE"
        fi
        ;;
    cleanup)
        if [ "${JS_CLEANUP_IMAGE}" = 1 ]
        then
          cleanup_image
        fi

        if [ "${JS_CLEANUP_JOURNAL}" = 1 ]
        then
          cleanup_journal
        fi

        if [ "${JS_CLEANUP_LOGS}" ] && [ "${JS_CONTROLLER}" = 1 ]
        then
          cleanup_logs controller "$JS_CONTROLLER_IMAGE"
        fi

        if [ "${JS_CLEANUP_LOGS}" ] && [ "${JS_AGENT}" = 1 ]
        then
          cleanup_logs agent "$JS_AGENT_IMAGE"
        fi

        if [ "${JS_CLEANUP_LOGS}" ] && [ "${JS_JOC}" = 1 ]
        then
          cleanup_logs joc "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CLEANUP_LOGS}" ] && [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          cleanup_logs controller "$JS_CONTROLLER_IMAGE"
          cleanup_logs agent      "$JS_AGENT_IMAGE"
          cleanup_logs joc        "$JS_JOC_IMAGE"
        fi
        ;;
diagnostic)
        if [ 1 = 0 ] && [ ! "$JS_DIAGNOSTIC_FILE_DESCRIPTORS" = 1 ]
        then
          echo "command 'diagnostic' requires -fd switch"
          exit 1
        fi

        if [ "$JS_CONTROLLER" = 0 ] && [ "$JS_AGENT" = 0 ] && [ "$JS_JOC" = 0 ]
        then
          is_denied=1
        fi

        if [ "$JS_CONTROLLER" = 1 ] && [[ $JS_CONTROLLER_IMAGE =~ ^\*.* ]]
        then
          is_denied=1
        fi

        if [ "$JS_AGENT" = 1 ] && [[ $JS_AGENT_IMAGE =~ ^\*.* ]]
        then
          is_denied=1
        fi

        if [ "$JS_JOC" = 1 ] && [[ $JS_JOC_IMAGE =~ ^\*.* ]]
        then
          is_denied=1
        fi

        if [ "$is_denied" = 1 ]
        then
          echo "command 'diagnostic' requires to select a container with either --controller=<container>, --agent=<container> or --joc=<container>"
          exit 1
        fi
        
        while true; do
          if [ "${JS_CONTROLLER}" = 1 ]
          then
            diagnostic_file_descriptors controller "$JS_CONTROLLER_IMAGE"
          fi

          if [ "${JS_AGENT}" = 1 ]
          then
            diagnostic_file_descriptors agent "$JS_AGENT_IMAGE"
          fi

          if [ "${JS_JOC}" = 1 ]
          then
            diagnostic_file_descriptors joc "$JS_JOC_IMAGE"
          fi

          sleep $JS_DIAGNOSTIC_REPEAT
        done
        ;;
   deploy)
        if [ "${JS_CONTROLLER}" = 1 ]
        then
          deploy controller "$JS_CONTROLLER_IMAGE"
        fi

        if [ "${JS_AGENT}" = 1 ]
        then
          deploy agent "$JS_AGENT_IMAGE"
        fi

        if [ "${JS_JOC}" = 1 ]
        then
          deploy joc "$JS_JOC_IMAGE"
        fi

        if [ "${JS_CONTROLLER}" = 0 ] && [ "${JS_AGENT}" = 0 ] && [ "${JS_JOC}" = 0 ]
        then
          deploy controller "$JS_CONTROLLER_IMAGE"
          deploy agent      "$JS_AGENT_IMAGE"
          deploy joc        "$JS_JOC_IMAGE"
        fi
        ;;
   submit)
        submit_order "$JS_SUBMIT_DIRECTORY" "$JS_SUBMIT_FILE" "$JS_SUBMIT_ID" "$JS_SUBMIT_COUNT" "$JS_SUBMIT_URL"
        ;;
   *)
        usage
        ;;
esac

